
# Identify the resource files for the bumps GUI
from ..resfiles import Resources
resources = Resources(package=__name__,
                      patterns=('*.png', '*.jpg', '*.ico', '*.wav'),
                      datadir='bumps-data',
                      check_file='reload.png',
                      env='BUMPS_DATA')
del Resources



import io
import csv
from typing import Literal, TypedDict, Any

def csv2dict(csvdata: str) -> dict:
    with io.StringIO(csvdata) as f:
        reader = csv.DictReader(f)
        result=dict(header=reader.fieldnames,
                    rows=[row for row in reader],
                    raw=csvdata)    
    
    return result

def dict2csv(csvdict: dict) -> str:
    # No longer needed since raw csv data are added to the table data
    # dictionary in csv2dict
    with io.StringIO() as f:
        writer = csv.DictWriter(f, fieldnames=csvdict['header'])
        writer.writerows(csvdict['rows'])
        result = f.getvalue()

    return result

class CustomWebviewPlot(TypedDict):
    fig_type: Literal['plotly', 'matplotlib', 'table', 'error'] = 'plotly'
    plotdata: Any

def process_custom_plot(plot_item: CustomWebviewPlot) -> CustomWebviewPlot:
    
    figtype = plot_item['fig_type']
    plot_data = plot_item['plotdata']

    if figtype == 'plotly':
        figdict = plot_data.to_dict()
    elif figtype == 'matplotlib':
        import mpld3
        figdict = mpld3.fig_to_dict(plot_data)
    elif figtype == 'table':
        figdict = csv2dict(plot_data)
    elif figtype == 'error':
        figdict = plot_data
    else:
        figdict = dict(error=f'unrecognized plot type {figtype}')
    
    del plot_data # is this necessary? Does plot_item['plot_data'] also need to be deleted?
    
    return CustomWebviewPlot(fig_type=figtype,
                             plotdata=figdict)




from __future__ import with_statement

import numpy as np

from .. import monitor
from ..plotutil import coordinated_colors
from .plot_view import PlotView


class ConvergenceMonitor(monitor.Monitor):
    """
    Gather statistics about the best, worst, median and +/- 1 interquartile
    range.  This will be the input for the convergence plot.
    """
    def __init__(self):
        self.pop = []
    def config_history(self, history):
        history.requires(population_values=1, value=1)
    def __call__(self, history):
        best = history.value[0]
        try:
            pop = history.population_values[0]
            n = len(pop)
            p = np.sort(pop)
            QI,Qmid, = int(0.2*n),int(0.5*n)
            self.pop.append((best, p[0],p[QI],p[Qmid],p[-1-QI],p[-1]))
        except (AttributeError, TypeError):
            self.pop.append((best, ))
    def progress(self):
        if not self.pop:
            return dict(pop=np.empty((0,1),'d'))
        else:
            return dict(pop=np.array(self.pop))


class ConvergenceView(PlotView):
    title = "Convergence"
    def plot(self):
        if not self.plot_state: return
        pop,best = self.plot_state
        with self.pylab_interface as pylab:
            pylab.clf()
            ni,npop = pop.shape
            iternum = np.arange(1,ni+1)
            tail = int(0.25*ni)
            c = coordinated_colors(base=(0.4,0.8,0.2))
            if npop==5:
                pylab.fill_between(iternum[tail:], pop[tail:,1], pop[tail:,3],
                                   color=c['light'], label='_nolegend_')
                pylab.plot(iternum[tail:],pop[tail:,2],
                           label="80% range", color=c['base'])
                pylab.plot(iternum[tail:],pop[tail:,0],
                           label="_nolegend_", color=c['base'])
            pylab.plot(iternum[tail:], best[tail:], label="best",
                       color=c['dark'])
            pylab.xlabel('iteration number')
            pylab.ylabel('chisq')
            pylab.legend()
            #pylab.gca().set_yscale('log')
            pylab.draw()
    def update(self, best, pop):
        self.plot_state = pop, best
        self.plot()
    def OnFitProgress(self, event):
        if event.problem != self.model: return
        pop = 2*event.pop/self.model.dof
        self.update(pop[:, 0], pop[:, 1:])



"""
Build trace plots (plotly)
"""

from typing import Union

__all__ = ['plot_trace']

import numpy as np
from .colors import COLORS

# TODO: when minimum python version is 3.10, can use | to combine types 
def plot_trace(x: Union[list, np.ndarray], ys: Union[list, np.ndarray], label='', alpha=0.4):

    import plotly.graph_objs as go

    traces = []
    color_idx = 0
    for y in ys:
        color = COLORS[color_idx % len(COLORS)]
        traces.append(go.Scatter(x=x,
                                y=y,
                                mode='lines',
                                showlegend=False,
                                line=dict(color=color),
                                opacity=alpha))
        color_idx += 1

    fig = go.Figure(data=traces)

    fig.update_layout(
        template = 'plotly_white',
        xaxis_title=dict(text='Generation number'),
        yaxis_title=dict(text=label),
        yaxis=dict(exponentformat='e'),
        showlegend=False
    )

    return fig


"""
2d array representation of a circle

References::
    None
"""

import numpy as np
from numpy import arange
from numpy import random, sin, cos, pi, inf, sqrt

random.seed(123)

def circle(coeffs,interval=0.02):
    """
    generate a 2D array representation of a circle of given coeffs
    coeffs = (x,y,r)
    """
    xc,yc,r = coeffs
    theta = arange(0, 2*pi, interval)
    return r*cos(theta)+xc, r*sin(theta)+yc

def genpoints(coeffs, npts=20):
    """
    Generate a 2D dataset of npts enclosed in circle of given coeffs,
    where coeffs = (x,y,r).

    NOTE: if npts is None, constrain all points to circle of given radius
    """
    xo,yo,R = coeffs
    # Radial density varies as sqrt(x)
    r,theta = sqrt(random.rand(npts)), 2*pi*(random.rand(npts))
    x,y = r*cos(theta)+xo, r*sin(theta)+yo
    return x,y

def gendensity(coeffs,density=0.1):
    xo,yo,R = coeffs
    npts = int(0.2*pi*R**2)
    return genpoints(coeffs,npts)

from .model import Fitness
class MinimumCircle(Fitness):
    def __init__(self, data=None, limits=None, start=None):
        self.x,self.y = data
        self.dy = None
        self.limits = limits
        self.start = start

    def _residuals(self, p):
        x,y = self.x,self.y
        xc,yc,r = p
        d = sqrt((x-xc)**2 + (y-yc**2))
        d[d<r] = 0
        return d

    def profile(self, p):
        return circle(p)

    def residuals(self, p):
        """
        Residuals is used by Levenburg-Marquardt, so fake the
        penalty terms of the normal cost function for use here.
        """
        resid = self._residuals(p)
        # Throw r in the residual so that it is minimized, punish the circle
        # if there are too many points outside.
        d = np.concatenate((resid,self.r,sum(resid>0)))
        return d

    def __call__(self, p):
        xc,yc,r = p
        resid = self._residuals(p)
        # Penalties are the number
        # Add additional penalties for each point outside the circle
        return sum(resid>0) + sum(resid) + abs(r)

# prepared instances
Po = [0,0,1]
Plo=[-inf,-inf,0]
Phi=[inf,inf,inf]
dense_circle = MinimumCircle(data=gendensity([3,4,20],density=5),
                             limits=(Plo,Phi), start=Po)
sparse_circle = MinimumCircle(data=gendensity([3,-4,32],density=0.5),
                              limits=(Plo,Phi), start=Po)
minimal_circle = MinimumCircle(data=gendensity([0,0,10],density=5),
                               limits=(Plo,Phi), start=Po)



# persist settings to disk

import sys
from pathlib import Path
from typing import Optional

ETC_PATH = Path(sys.prefix, 'etc')
DEFAULT_APPLICATION = 'bumps'

def set_value(key: str, value: str, application: str = DEFAULT_APPLICATION):
    settings_path = ETC_PATH / application / key
    try:
        settings_path.parent.mkdir(parents=True, exist_ok=True)
        with open(settings_path, 'w') as f:
            f.write(value)
        return True
    except Exception as e:
        return False

def get_value(key: str, default_value: Optional[str] = None, application: str = DEFAULT_APPLICATION):
    settings_path = ETC_PATH / application / key
    try:
        with open(settings_path, 'r') as f:
            return f.read()
    except Exception as e:
        return default_value



"""
Ccorana's function

References::
    [1] Storn, R. and Price, K. Differential Evolution - A Simple and Efficient
    Heuristic for Global Optimization over Continuous Spaces. Journal of Global
    Optimization 11: 341-359, 1997.

    [2] Storn, R. and Price, K.
    (Same title as above, but as a technical report.)
    http://www.icsi.berkeley.edu/~storn/deshort1.ps
"""
from math import pow

from numpy import sign, floor, inf

def corana(coeffs):
    """
    evaluates the Corana function for a list of coeffs

    minimum is f(x)=0.0 at xi=0.0
    """
    d = [1., 1000., 10., 100.]
    x = coeffs
    r = 0
    for xj,dj in zip(x,d):
        zj =  floor( abs(xj/0.2) + 0.49999 ) * sign(xj) * 0.2
        if abs(xj-zj) < 0.05:
            r += 0.15 * pow(zj - 0.05*sign(zj), 2) * dj
        else:
            r += dj * xj ** 2
        return r


def corana1d(x):
    """Corana in 1D; coeffs = (x,0,0,0)"""
    return corana([x[0], 0, 0, 0])

def corana2d(x):
    """Corana in 2D; coeffs = (x,0,y,0)"""
    return corana([x[0], 0, x[1], 0])

def corana3d(x):
    """Corana in 3D; coeffs = (x,0,y,z)"""
    return corana([x[0], 0, x[1], x[2]])

Po = [1,1,1,1]
Plo = [-inf,-inf,-inf,-inf]
Phi = [inf,inf,inf,inf]

def s(V,*args):
    retval = []
    for i in args: retval.append(V[i])
    return retval
from .model import Function
corana1d = Function(f=corana1d, limits=(s(Plo,0),s(Phi,0)),
                    start=s(Po,0))
corana2d = Function(f=corana2d, limits=(s(Plo,0,2),s(Phi,0,2)),
                    start=s(Po,0,2))
corana3d = Function(f=corana3d, limits=(s(Plo,0,2,3),s(Phi,0,2,3)),
                    start=s(Po,0,2,3))
corana4d = Function(f=corana, limits=(Plo,Phi), start=Po)
# End of file
